#!/bin/csh -f
#
# ccombine.csh - CARMA STING track combination script (Tony Wong, UIUC)
#
# History
#  2009aug10: Initial version
#--------------------------------------------------------------------------

# Basic params
set files=
set source=
set interact=y

# Mapping and deconvolution params
set vfirst=         # velocity of first channel
set nmaps=          # number of maps to generate
set delv=           # channel spacing in km/s
set imsize=513      # for INVERT
set cell=.5         # for INVERT
set exinvert=       # extra parameters for INVERT
set rmsfac=3        # rmsfac for mossdi
set clnbox=         # deconvoltion region, whole image if unset
set bmaj=           # beam major axis, arcsec
set bmin=           # beam minor axis, arcsec
set bpa=            # beam position angle, deg

# Moment params
set fwhm=12         # FWHM of convolving Gaussian in smoothed mask method
set mom0fun=sqr     # Use n^2 (sqr) or logarithmic (log) contours from mom-0 plots
set mom0lev=        # First contour level, default is 3-sigma
set loglevs=0.2     # logarithmic contour level spacing in dex
set mom1del=20      # Spacing between velocity contours, in km/s
set fluxclip=4      # min allowed flux (Jy/bm km/s) for gaufits
set velclip=20      # max allowed velocity error (km/s) for gaufits

### LOCAL SETTINGS
alias date 'date "+%m/%d/%y %T %Z"'
set version=2009aug10
set log=ccombine.log

### OVERRIDE ABOVE PARAMETERS IF GIVEN ON COMMAND LINE
foreach par ( $* )
   set check=`echo $par | awk -F= '{print NF}'`
   if ( "$check" >= 2 ) set $par
end

#============================================================================

# OPEN LOGFILE IF NEEDED
echo ""
if (-e $log && $interact != 'n') then
    echo -n "$log exists.  Overwrite (o) or append (a) [default=a]? "
    set resp = $<
    if ($resp == "o") rm -f $log
    echo ""
endif

# BEGIN EXECUTION
goto continue
# MOVE 'continue:' around to where you want to work.
# Shouldn't be necessary in standard operation.
continue:

# MAIN MENU
if !($?start) then
    echo "Where would you like to start?"
    echo "  1. Make dirty maps and plot"
    echo "  2. Deconvolve maps and plot"
    echo "  3. Make moment maps"
    echo "  4. Make Gaussian fit velocity field"
    echo "  5. Clean up"
    echo ""
    echo -n "? "
    set start = $<
endif
switch ($start)
    case 2: 
    goto deconv
    case 3: 
    goto domom
    case 4: 
    goto dogauss
    case 5: 
    goto dorm
endsw

# Output some interesting information
if !(-e $log) then
    echo ""
    echo "ccombine.csh $version executed by $user on `date`" | tee $log
    echo "" | tee -a $log
endif

#--------------------------------------------------------------------------
# 1. Make dirty map:
#--------------------------------------------------------------------------

# Make the map
rm -rf $source.map $source.beam
echo "Processing files: $files"
echo "Making $nmaps maps beginning at $vfirst and spaced by $delv km/s" 
invert vis="$files" map=$source.map beam=$source.beam cell=$cell \
    imsize=$imsize line=velocity,$nmaps,$vfirst,$delv,$delv \
    options=systemp,double,mosaic slop=1 robust=0.5 $exinvert
    
# Calculate rms
set nmin1 = `expr $nmaps - 1`
echo "Dirty map, Channels 1-2, $nmin1-$nmaps, Mean:" | tee -a $log
set exec="histo in=$source.map"
set box="box(-60,-60,60,60)"
set crmsi=`$exec region="a,$box(1,2)" |grep Rms |awk '{print $4}'`
set crmsf=`$exec region="a,$box($nmin1,$nmaps)" |grep Rms |awk '{print $4}'`
set crms = `calc -f f7.5 "($crmsi+$crmsf)/2"`
echo "RMS in central 2 x 2 arcm: $crmsi $crmsf  Mean: $crms" | tee -a $log
set frmsi=`$exec region="image(1,2)" | grep Rms | awk '{print $4}'`
set frmsf=`$exec region="image($nmin1,$nmaps)" | grep Rms | awk '{print $4}'`
set frms = `calc -f f7.5 "($frmsi+$frmsf)/2"`
echo "RMS in entire field: $frmsi $frmsf  Mean: $frms" | tee -a $log

# Get the (flattened) noise level.
rm -rf $source.sen
mossen in=$source.map sen=$source.sen
rm -rf $source.mapnorm
set minval=`histo in=$source.sen region='image(1)'|grep Min|awk '{print $3}'`
echo "Normalizing $source.sen by $minval" | tee -a $log
maths exp="$source.map*$minval/$source.sen" out=$source.mapnorm
set clip=`calc "2*$minval"`
rm -rf $source.senmsk
maths exp=$source.sen out=$source.senmsk mask="<$source.sen>.lt.$clip"
set nmin1 = `expr $nmaps - 1`
set exec="histo in=$source.mapnorm"
set box="box(-60,-60,60,60)"
set crmsi=`$exec region="a,$box(1,2)" |grep Rms |awk '{print $4}'`
set crmsf=`$exec region="a,$box($nmin1,$nmaps)"|grep Rms|awk '{print $4}'`
set crms = `calc -f f7.5 "($crmsi+$crmsf)/2"`
echo "RMS noise at center: $crms" | tee -a $log
set scale=`calc -f f7.3 "$crms/$minval"`
echo "Scaling factor for rmsfac: $scale" | tee -a $log
puthd in=$source.mapnorm/rms value=$crms type=real
puthd in=$source.mapnorm/scale value=$scale type=real
echo " "

# Make the plot.
cgdisp in=$source.mapnorm device=$source.dmaps.ps/ps \
	region='arc,box(-90,-90,90,90)' \
    options=full,3value labtyp=arcsec nxy=4,3 type=contour \
    slev=a,$crms levs1=2,3,4,5,6,8,10,12,14 csize=0.6,0.8
gv $source.dmaps.ps &

if ($interact != 'n') then
	echo -n "Continue with deconvolution [n]? "
	set resp=$<
	if ($resp != 'y') goto end
endif

#--------------------------------------------------------------------------
# 2. Clean the map:
#--------------------------------------------------------------------------

deconv:
# Deconvolve using MOSSDI
set scale=`gethd in=$source.mapnorm/scale`
set cutoff=`calc "$scale*$rmsfac"`
echo "MOSSDI2: rmsfac=$rmsfac, cutoff=$cutoff" | tee -a $log
rm -rf $source.sdi
if ($clnbox == '') then
	mossdi2 map=$source.map beam=$source.beam out=$source.sdi cutoff=$cutoff \
		niters=20000
else
	mossdi2 map=$source.map beam=$source.beam out=$source.sdi cutoff=$cutoff \
		niters=20000 region=$clnbox
endif
rm -rf $source.psf
mospsf beam=$source.beam out=$source.psf
if ($bmaj == '') then
    imfit in=$source.psf object=beam region="a,box(-10,-10,10,10)" \
        | tee tmp.imfit
    set bmaj=`grep 'Major axis' tmp.imfit | sed 's/.*://' | awk '{print $1}'`
    set bmin=`grep 'Minor axis' tmp.imfit | sed 's/.*://' | awk '{print $1}'`
    set bpa=`grep 'Position angle' tmp.imfit | sed 's/.*://' | awk '{print $1}'`
    rm -f tmp.imfit
else
    if ($bmin == '') set bmin=$bmaj
    if ($bpa == '') set bpa=0
endif
rm -rf $source.cm
restor map=$source.map beam=$source.beam model=$source.sdi out=$source.cm \
    fwhm=$bmaj,$bmin pa=$bpa
rm -rf $source.res
restor map=$source.map beam=$source.beam model=$source.sdi out=$source.res \
    fwhm=$bmaj,$bmin pa=$bpa mode=residual
copyhd in=$source.map out=$source.cm items=mask
set minval=`histo in=$source.sen region='image(1)'|grep Min|awk '{print $3}'`
rm -rf $source.cmnorm
maths exp="$source.cm*$minval/$source.sen" out=$source.cmnorm
rm -rf $source.resnorm
maths exp="$source.res*$minval/$source.sen" out=$source.resnorm
copyhd in=$source.cm out=$source.res items=bmaj,bmin,bpa
set rflux=`histo in=$source.res region="a,box(-60,-60,60,60)" | grep Flux | awk '{print $6}'`
set rintflux=`calc -f f8.2 "$rflux*$delv"`
set cflux=`histo in=$source.sdi | grep Flux | awk '{print $6}'`
set cintflux=`calc -f f8.2 "$cflux*$delv"`
echo "Used restoring beam of $bmaj x $bmin arcsec, pa $bpa" | tee -a $log
echo "Flux of residuals in region: $rintflux Jy km/s" | tee -a $log
echo "Flux of deconv. model in region: $cintflux Jy km/s" | tee -a $log

set crms=`gethd in=$source.mapnorm/rms`
set maxval=`histo in=$source.cmnorm | grep Max | awk '{print $3}'`
cgdisp in=$source.cmnorm,$source.cmnorm device=$source.cmaps.ps/ps \
    region='arc,box(-90,-90,90,90)' options=full,3value,mirr \
    type=c,p labtyp=arcsec nxy=4,3 lines=1,1,2 slev=a,$crms \
    levs1=2,3,4,5,6,8,10,12,14 csize=0.6,0.8 range=0,$maxval,lin,1
gv $source.cmaps.ps &

if ($interact != 'n') then
	echo -n "Continue with moment maps [n]? "
	set resp=$<
	if ($resp != 'y') goto end
endif

#--------------------------------------------------------------------------
# 3. Make moment maps:
#--------------------------------------------------------------------------

domom:
# Make moment maps
set crms=`gethd in=$source.mapnorm/rms`
moment.csh file=$source.cm rms=$crms noisemap=$source.sen start=straight \
	clip1pk=3 interact=$interact
moment.csh file=$source.cm rms=$crms noisemap=$source.sen start=mask \
	fwhm=$fwhm interact=$interact

# Plot the mom-0 maps
mv $source.mmom0/mask $source.mmom0/mask1
set maxval=`histo in=$source.mmom0 | grep Max | awk '{print $3}'`
if ($mom0fun == 'sqr') then
	set levs=1,4,9,16,25,36,49,81,100
else
	set levs=`awk -v del=$loglevs 'BEGIN {for (i=0;i<10;i++) printf "%3.1f,",10**(i*del); printf "%3.1f\n",10**(10*del)}'`
endif
set sig3=`calc "3*$crms*sqrt($nmaps)*$delv"`
echo "The 3-sigma level is $sig3"
if ($mom0lev == '') then
	set mom0lev=$sig3
endif
cgdisp in=$source.mmom0,$source.mmom0,$source.nse type=p,c,c \
    region='arc,box(-90,-90,90,90)(1)' device=$source.mmom0.ps/cps \
    range=0,$maxval,lin,-8 slev=a,$mom0lev,a,1 levs1=$levs levs2=2 \
	labtyp=hms,dms nxy=1,1 options=black,solneg2,wedge,full cols1=1 csize=1 \
	lines=2,1,2 beamtyp=b,l
gv $source.mmom0.ps &
mv $source.mmom0/mask1 $source.mmom0/mask

# Plot the mom-1 maps
set vlast=`calc "$vfirst+($nmaps-1)*$delv"`
set nlevs=`calc -i "($vlast-$vfirst)/$mom1del"`
awk -v v0=$vfirst -v dv=$mom1del -v nl=$nlevs \
	'BEGIN { for (i = 1; i < nl; i++) {printf "%d,",v0+i*dv} printf "%d\n",v0+nl*dv}' \
	> vlevs.txt
cgdisp in=$source.vpeak,$source.vpeak type=p,c \
    region='arc,box(-90,-90,90,90)' device=$source.vpeak.ps/cps \
    range=$vfirst,$vlast,lin,3 slev=a,1 levs1=@vlevs.txt \
	labtyp=hms,dms nxy=1,1 options=black,wedge,full cols1=1 csize=1 lines=2,1
gv $source.vpeak.ps &

if ($interact != 'n') then
	echo -n "Continue with Gaussian fits [n]? "
	set resp=$<
	if ($resp != 'y') goto end
endif

#--------------------------------------------------------------------------
# 4. Make Gaussfit velocity map:
#--------------------------------------------------------------------------

dogauss:
set crms=`gethd in=$source.mapnorm/rms`
moment.csh file=$source.cm rms=$crms noisemap=$source.sen start=gaufit \
	fluxclip=$fluxclip velclip=$velclip

# Plot the mom-1 maps
set vlast=`calc "$vfirst+($nmaps-1)*$delv"`
set nlevs=`calc -i "($vlast-$vfirst)/$mom1del"`
awk -v v0=$vfirst -v dv=$mom1del -v nl=$nlevs \
	'BEGIN { for (i = 1; i < nl; i++) {printf "%d,",v0+i*dv} printf "%d\n",v0+nl*dv}' \
	> vlevs.txt
cgdisp in=$source.gmom1,$source.gmom1 type=p,c \
    region='arc,box(-90,-90,90,90)' device=$source.gvel.ps/cps \
    range=$vfirst,$vlast,lin,3 slev=a,1 levs1=@vlevs.txt \
	labtyp=hms,dms nxy=1,1 options=black,wedge,full cols1=1 csize=1 lines=2,1
gv $source.gvel.ps &

if ($interact != 'n') then
	echo -n "Continue with cleanup [n]? "
	set resp=$<
	if ($resp != 'y') goto end
else
	goto end
endif

#--------------------------------------------------------------------------
# 5. Clean up files
#--------------------------------------------------------------------------

dorm:
# Clean up:
moment.csh file=$source.cm start=cleanup
rm -rf $source.res $source.resnorm $source.psf
rm -rf $source.sen $source.senmsk $source.gau*
gzip -f *.ps

end:

